#include <QtGui>
#include "QHexEdit_p.h"


const int HEXCHARS_IN_LINE = 47;
const int GAP_ADR_HEX = 10;
const int GAP_HEX_ASCII = 16;
const int BYTES_PER_LINE = 16;


QHexEditPrivate::QHexEditPrivate(QScrollArea *parent) : QWidget(parent)
{
	_scrollArea = parent;
	setAddressWidth(4);
	setAddressOffset(0);
	setAddressArea(true);
	setAsciiArea(true);
	setHighlighting(true);
	setAddressAreaColor(QColor(Qt::lightGray).lighter(110));
	setHighlightingColor(QColor(Qt::yellow).lighter(160));

	QFont font("Courier New", 10);
	font.setStyleHint(QFont::Monospace);
	setFont(font);
	connect(&_cursorTimer, SIGNAL(timeout()), SLOT(updateCursor()));

	_cursorTimer.setInterval(500);
	_cursorTimer.start();

	setFocusPolicy(Qt::StrongFocus);
	_size = -1;
}


void QHexEditPrivate::setAddressOffset(int offset)
{
	_addressOffset = offset;
	adjust();
}


int QHexEditPrivate::addressOffset()
{
	return _addressOffset;
}


void QHexEditPrivate::setData(const QByteArray &data)
{
	_data = data;
	_originalData = data;
	adjust();
	setCursorPos(0);
	setFocus();
}


QByteArray QHexEditPrivate::data()
{
	return _data;
}


void QHexEditPrivate::setAddressAreaColor(const QColor &color)
{
	_addressAreaColor = color;
	update();
}


QColor QHexEditPrivate::addressAreaColor()
{
	return _addressAreaColor;
}


void QHexEditPrivate::setHighlightingColor(const QColor &color)
{
	_highlightingColor = color;
	update();
}


QColor QHexEditPrivate::highlightingColor()
{
	return _highlightingColor;
}


void QHexEditPrivate::setAddressArea(bool addressArea)
{
	_addressArea = addressArea;
	adjust();
	setCursorPos(_cursorPosition);
}


void QHexEditPrivate::setAddressWidth(int addressWidth)
{
	if ((addressWidth >= 0) and (addressWidth<=6))
	{
		_addressNumbers = addressWidth;
		adjust();
		setCursorPos(_cursorPosition);
	}
}


void QHexEditPrivate::setAsciiArea(bool asciiArea)
{
	_asciiArea = asciiArea;
	adjust();
}


void QHexEditPrivate::setFont(const QFont &font)
{
	QWidget::setFont(font);
	adjust();
}


void QHexEditPrivate::setHighlighting(bool mode)
{
	_highlighting = mode;
	update();
}


void QHexEditPrivate::keyPressEvent(QKeyEvent *event)
{
	bool down = false;
	int charX = (_cursorX - _xPosHex) / _charWidth;
	int posX = (charX / 3) * 2 + (charX % 3);
	int posBa = (_cursorY / _charHeight) * BYTES_PER_LINE + posX / 2;


	int key = int(event->text()[0].toAscii());
	if ((key>='0' && key<='9') || (key>='a' && key<='f') || (key>='A' && key<='F'))
	{
		if (_data.size() > 0)
		{
			QByteArray hexValue = _data.mid(posBa, 1).toHex();
			if ((charX % 3) == 0)
				hexValue[0] = key;
			else
				hexValue[1] = key;
			_data.replace(posBa, 1, QByteArray().fromHex(hexValue));
			emit dataChanged();


			setCursorPos(_cursorPosition + 1);
			down = true;
		}
	}


	if (event->matches(QKeySequence::MoveToNextChar))
	{
		setCursorPos(_cursorPosition + 1);
		down = true;
	}
	if (event->matches(QKeySequence::MoveToPreviousChar))
		setCursorPos(_cursorPosition - 1);
	if (event->matches(QKeySequence::MoveToStartOfLine))
		setCursorPos(_cursorPosition - (_cursorPosition % (2 * BYTES_PER_LINE)));
	if (event->matches(QKeySequence::MoveToEndOfLine))
		setCursorPos(_cursorPosition | (2 * BYTES_PER_LINE -1));
	if (event->matches(QKeySequence::MoveToPreviousLine))
		setCursorPos(_cursorPosition - (2 * BYTES_PER_LINE));
	if (event->matches(QKeySequence::MoveToPreviousPage))
		setCursorPos(_cursorPosition - (((_scrollArea->viewport()->height() / _charHeight) - 1) * 2 * BYTES_PER_LINE));
	if (event->matches(QKeySequence::MoveToStartOfDocument))
		setCursorPos(0);
	if (event->matches(QKeySequence::MoveToNextLine))
	{
		setCursorPos(_cursorPosition + (2 * BYTES_PER_LINE));
		down = true;
	}
	if (event->matches(QKeySequence::MoveToEndOfDocument))
	{
		setCursorPos(_data.size() * 2);
		down = true;
	}
	if (event->matches(QKeySequence::MoveToNextPage))
	{
		setCursorPos(_cursorPosition + (((_scrollArea->viewport()->height() / _charHeight) - 1) * 2 * BYTES_PER_LINE));
		down = true;
	}


	// when we move downwards, we have to go a little further
	if (down)
		_scrollArea->ensureVisible(_cursorX, _cursorY, 3, 3 + _charHeight);
	else
		_scrollArea->ensureVisible(_cursorX, _cursorY, 3, 3);
	update();
}


void QHexEditPrivate::mousePressEvent(QMouseEvent * event)
{
	setCursorPos(event->pos());
}


void QHexEditPrivate::paintEvent(QPaintEvent *event)
{
	QPainter painter(this);


	// draw some patterns if needed
	painter.fillRect(event->rect(), this->palette().color(QPalette::Base));
	if (_addressArea)
		painter.fillRect(QRect(_xPosAdr, event->rect().top(), _xPosHex - GAP_ADR_HEX + 2, height()), _addressAreaColor);
	if (_asciiArea)
	{
		int linePos = _xPosAscii - (GAP_HEX_ASCII / 2);
		painter.setPen(Qt::gray);
		painter.drawLine(linePos, event->rect().top(), linePos, height());
	}


	painter.setPen(this->palette().color(QPalette::Text));


	// calc position
	int firstLineIdx = ((event->rect().top()/ _charHeight) - _charHeight) * BYTES_PER_LINE;
	if (firstLineIdx < 0)
		firstLineIdx = 0;
	int lastLineIdx = ((event->rect().bottom() / _charHeight) + _charHeight) * BYTES_PER_LINE;
	if (lastLineIdx > _data.size())
		lastLineIdx = _data.size();
	int yPosStart = ((firstLineIdx) / BYTES_PER_LINE) * _charHeight + _charHeight;


	// paint address area
	if (_addressArea)
	{
		for (int lineIdx = firstLineIdx, yPos = yPosStart; lineIdx < lastLineIdx; lineIdx += BYTES_PER_LINE, yPos +=_charHeight)
		{
			painter.drawText(_xPosAdr, yPos, QString("%1")
							 .arg(lineIdx + _addressOffset, _realAddressNumbers, 16, QChar('0')));
		}
	}


	// paint hex area
	QByteArray hexBa(_data.mid(firstLineIdx, lastLineIdx - firstLineIdx + 1).toHex());
	painter.setBackground(QBrush(_highlightingColor));
	painter.setBackgroundMode(Qt::TransparentMode);
	for (int lineIdx = firstLineIdx, yPos = yPosStart; lineIdx < lastLineIdx; lineIdx += BYTES_PER_LINE, yPos +=_charHeight)
	{
		int xPos = _xPosHex;
		for (int colIdx = 0; ((lineIdx + colIdx) < _data.size() and (colIdx < BYTES_PER_LINE)); colIdx++)
		{
			// hilight diff bytes
			if (_highlighting)
			{
				int posBa = lineIdx + colIdx;
				if (posBa >= _originalData.size())
					painter.setBackgroundMode(Qt::TransparentMode);
				else if (_data[posBa] == _originalData[posBa])
					painter.setBackgroundMode(Qt::TransparentMode);
				else
					painter.setBackgroundMode(Qt::OpaqueMode);
			}


			// render hex value
			if (colIdx == 0)
			{
				painter.drawText(xPos, yPos, QString(hexBa.at((lineIdx - firstLineIdx) * 2)));
				xPos += _charWidth;
				painter.drawText(xPos, yPos, QString(hexBa.at((lineIdx - firstLineIdx) * 2 + 1)));
				xPos += _charWidth;
			}
			else
			{
				painter.drawText(xPos, yPos, " ");
				xPos += _charWidth;
				painter.drawText(xPos, yPos, QString(hexBa.at((lineIdx + colIdx - firstLineIdx) * 2)));
				xPos += _charWidth;
				painter.drawText(xPos, yPos, QString(hexBa.at((lineIdx + colIdx - firstLineIdx) * 2 + 1)));
				xPos += _charWidth;
			}
		}
	}
	painter.setBackgroundMode(Qt::TransparentMode);


	// paint ascii area
	if (_asciiArea)
	{
		for (int lineIdx = firstLineIdx, yPos = yPosStart; lineIdx < lastLineIdx; lineIdx += BYTES_PER_LINE, yPos +=_charHeight)
		{
			QByteArray ascii = _data.mid(lineIdx, BYTES_PER_LINE);
			for (int idx=0, xPos=_xPosAscii ; idx<ascii.size() ; ++idx, xPos += _charWidth) {
				char comp = ascii.at(idx);
				if ((quint8)comp < 0x20 || (quint8)comp > 0xe2)
					comp = '.';
				else
					comp = FF8Text::charFromff8(comp);

				// hilight current ascii
				if(_cursorPosition==(lineIdx+idx)*2 || _cursorPosition==(lineIdx+idx)*2+1)
				{
					painter.setBackground(QBrush(QColor(Qt::cyan).lighter(160)));
					painter.setBackgroundMode(Qt::OpaqueMode);
				}
				else
				{
					painter.setBackgroundMode(Qt::TransparentMode);
				}
				painter.drawText(xPos, yPos, QString(comp));
			}


		}
	}


	// paint cursor
	if (_blink)
	{
		painter.fillRect(_cursorX, _cursorY + _charHeight - 2, _charWidth, 2, this->palette().color(QPalette::WindowText));
//		painter.fillRect(_cursorX, _cursorY, _charWidth, _charHeight, this->palette().color(QPalette::WindowText));
	}


	if (_size != _data.size())
	{
		_size = _data.size();
		emit currentSizeChanged(_size);
	}
}


void QHexEditPrivate::setCursorPos(int position)
{
	// delete cursor
	_blink = false;
	update();

	// cursor in range?
	if (position > (_data.size() * 2 - 1))
		position = _data.size() * 2 - 1;


	if (position < 0)
		position = 0;


	// calc position
	_cursorPosition = position;
	_cursorY = (position / (2 * BYTES_PER_LINE)) * _charHeight + 4;
	int x = (position % (2 * BYTES_PER_LINE));
	_cursorX = (((x / 2) * 3) + (x % 2)) * _charWidth + _xPosHex;


	// immiadately draw cursor
	_blink = true;
	update();
	emit currentAddressChanged(_cursorPosition/2);
}


void QHexEditPrivate::setCursorPos(QPoint pos)
{
	// find char under cursor
	if ((pos.x() >= _xPosHex) and (pos.x() < (_xPosHex + HEXCHARS_IN_LINE * _charWidth)))
	{
		int x = (pos.x() - _xPosHex) / _charWidth;
		if ((x % 3) == 0)
			x = (x / 3) * 2;
		else
			x = ((x / 3) * 2) + 1;
		int y = (pos.y() / _charHeight) * 2 * BYTES_PER_LINE;
		setCursorPos(x + y);
	}
}


void QHexEditPrivate::updateCursor()
{
	_blink = !_blink;

	update(_cursorX, _cursorY, _charWidth, _charHeight);
}


void QHexEditPrivate::adjust()
{
	_charWidth = fontMetrics().width(QChar('M'));
	_charHeight = fontMetrics().height();


	// is addressNumbers wide enought?
	_realAddressNumbers = QString("%1")
						  .arg(_data.size() + _addressOffset, _addressNumbers, 16, QChar('0')).size();


	_xPosAdr = 0;
	if (_addressArea)
		_xPosHex = _realAddressNumbers *_charWidth + GAP_ADR_HEX;
	else
		_xPosHex = 0;
	_xPosAscii = _xPosHex + HEXCHARS_IN_LINE * _charWidth + GAP_HEX_ASCII;


	// tell QAbstractScollbar, how big we are
	setMinimumHeight(((_data.size()/16 + 1) * _charHeight) + 3);
	setMinimumWidth(_xPosAscii + (BYTES_PER_LINE * _charWidth));


	update();
}
